﻿using System;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// 메인 폼
    /// </summary>
    public partial class MainForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Import
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 비트 복사하기 - BitBlt(targetHandle, targetX, targetY, targetWidth, targetHeight, sourceHandle, sourceX, sourceY, rasterOperation)

        /// <summary>
        /// 비트 복사하기
        /// </summary>
        /// <param name="targetHandle">타겟 디바이스 컨텍스트 핸들</param>
        /// <param name="targetX">타겟 X</param>
        /// <param name="targetY">타겟 Y</param>
        /// <param name="targetWidth">타겟 너비</param>
        /// <param name="targetHeight">타겟 높이</param>
        /// <param name="sourceHandle">소스 디바이스 컨텍스트 핸들</param>
        /// <param name="sourceX">소스 X</param>
        /// <param name="sourceY">소스 Y</param>
        /// <param name="rasterOperation">래스터 작업</param>
        /// <returns>처리 결과</returns>
        [DllImport("gdi32.dll")]
        private static extern int BitBlt(IntPtr targetHandle, int targetX, int targetY, int targetWidth, int targetHeight, IntPtr sourceHandle, int sourceX, int sourceY, int rasterOperation);

        #endregion
        #region 객체 선택하기 - SelectObject(targetHandle, sourceObjectHandle)

        /// <summary>
        /// 객체 선택하기
        /// </summary>
        /// <param name="targetHandle">타겟 디바이스 컨텍스트 핸들</param>
        /// <param name="sourceObjectHandle">소스 객체 핸들</param>
        /// <returns>이전 객체</returns>
        [DllImport("gdi32.dll", EntryPoint = "SelectObject")]
        private static extern IntPtr SelectObject(IntPtr targetHandle, IntPtr sourceObjectHandle);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// SRC_COPY
        /// </summary>
        private const int SRC_COPY = 0xcc0020;

        /// <summary>
        /// 사각형 비트맵
        /// </summary>
        private Bitmap rectangleBitmap;

        /// <summary>
        /// X
        /// </summary>
        private int x = 50;

        /// <summary>
        /// Y
        /// </summary>
        private int y = 50;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainForm()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            SetStyle(ControlStyles.DoubleBuffer        , true);
            SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            SetStyle(ControlStyles.UserPaint           , true);

            Paint   += Form_Paint;
            KeyDown += Form_KeyDown;
            Load    += Form_Load;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 폼 로드시 처리하기 - Form_Load(sender, e)

        /// <summary>
        /// 폼 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Form_Load(object sender, EventArgs e)
        {
            using(Graphics graphics = CreateGraphics())
            {
                this.rectangleBitmap = new Bitmap(60, 60, graphics);

                using(Graphics bitmapGraphics = Graphics.FromImage(this.rectangleBitmap))
                {
                    bitmapGraphics.Clear(Color.White);

                    bitmapGraphics.DrawRectangle(Pens.Blue, 0, 0, 59, 59);
                }
            }
        }

        #endregion
        #region 폼 키 DOWN 처리하기 - Form_KeyDown(sender, e)

        /// <summary>
        /// 폼 키 DOWN 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Form_KeyDown(object sender, KeyEventArgs e)
        {
            if(e.KeyCode == Keys.Up)
            {
                this.y -= 10;
            }
            else if(e.KeyCode == Keys.Down)
            {
                this.y += 10;
            }
            else if(e.KeyCode == Keys.Left)
            {
                this.x -= 10;
            }
            else if(e.KeyCode == Keys.Right)
            {
                this.x += 10;
            }

            Invalidate();
        }

        #endregion
        #region 폼 PAINT 처리하기 - Form_Paint(sender, e)

        /// <summary>
        /// 폼 PAINT 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Form_Paint(object sender, PaintEventArgs e)
        {
            Graphics targetGraphics = e.Graphics;

            IntPtr targetDeviceContextHandle = targetGraphics.GetHdc();

            using(Graphics sourceGraphics = Graphics.FromImage(this.rectangleBitmap))
            {
                IntPtr sourceDeviceContextHandle = sourceGraphics.GetHdc();

                IntPtr previousObject = SelectObject(sourceDeviceContextHandle, this.rectangleBitmap.GetHbitmap());

                BitBlt(targetDeviceContextHandle, this.x, this.y, 60, 60, sourceDeviceContextHandle, 0, 0, SRC_COPY);

                sourceGraphics.ReleaseHdc(sourceDeviceContextHandle);

                if(previousObject != IntPtr.Zero)
                {
                    SelectObject(sourceDeviceContextHandle, previousObject);
                }
            }

            targetGraphics.ReleaseHdc(targetDeviceContextHandle);
        }

        #endregion
    }
}